import zipfile
from collections.abc import Sequence
from pathlib import Path

import pandas as pd
from biopsykit.signals.ecg.preprocessing import EcgPreprocessingNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass
from biopsykit.utils.dtypes import EcgRawDataFrame, IcgRawDataFrame
from biopsykit.utils.file_handling import get_subject_dirs

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.datasets._helper import compute_reference_heartbeats
from pepbench.utils._types import path_t

HERE = Path(__file__).parent
EXAMPLE_DATA_PATH = HERE.joinpath("../../../example_data")


class ExampleDataset(BasePepDatasetWithAnnotations):
    example_file_path: path_t

    def __init__(
        self,
        example_file_path: path_t,
        groupby_cols: Sequence[str] | None = None,
        subset_index: Sequence[str] | None = None,
        *,
        return_clean: bool = True,
    ) -> None:
        self.example_file_path = example_file_path
        # unzip the example dataset
        with zipfile.ZipFile(str(self.example_file_path)) as zf:
            zf.extractall(EXAMPLE_DATA_PATH)
        super().__init__(groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean)

    def create_index(self) -> pd.DataFrame:
        participant_ids = [
            participant_dir.name for participant_dir in get_subject_dirs(EXAMPLE_DATA_PATH, "VP_[0-9]{3}")
        ]

        index = pd.DataFrame(participant_ids, columns=["participant"])
        return index

    @property
    def sampling_rate_ecg(self) -> int:
        return 500

    @property
    def sampling_rate_icg(self) -> int:
        return 500

    @property
    def ecg(self) -> EcgRawDataFrame:
        if not self.is_single(None):
            raise ValueError("ECG data can only be accessed for a single participant and a single phase!")
        data = self._load_data("ecg")
        if self.return_clean:
            algo = EcgPreprocessingNeurokit()
            algo.clean(ecg=data, sampling_rate_hz=self.sampling_rate_ecg)
            return algo.ecg_clean_
        return data

    @property
    def icg(self) -> IcgRawDataFrame:
        if not self.is_single(None):
            raise ValueError("ICG data can only be accessed for a single participant and a single phase!")
        data = self._load_data("icg")
        if self.return_clean:
            algo = IcgPreprocessingBandpass()
            algo.clean(icg=data, sampling_rate_hz=self.sampling_rate_icg)
            return algo.icg_clean_
        return data

    def _load_data(self, data_type: str) -> pd.DataFrame:
        p_id = self.index["participant"][0]
        data = pd.read_csv(
            EXAMPLE_DATA_PATH.joinpath(f"{p_id}/{p_id.lower()}_{data_type}_data.csv"),
            index_col=0,
        )
        data.index = pd.DatetimeIndex(data.index)
        return data

    @property
    def reference_labels_ecg(self) -> pd.DataFrame:
        """Return the reference labels for the ECG signal.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ECG signal as a pandas DataFrame

        """
        return self._load_reference_labels("ECG")

    @property
    def reference_labels_icg(self) -> pd.DataFrame:
        """Return the reference labels for the ICG signal.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ICG signal as a pandas DataFrame

        """
        return self._load_reference_labels("ICG")

    def _load_reference_labels(self, channel: str) -> pd.DataFrame:
        participant = self.index["participant"][0]

        if not (self.is_single(None)):
            raise ValueError("Reference data can only be accessed for a single participant.")

        file_path = EXAMPLE_DATA_PATH.joinpath(
            f"{participant}/reference_labels/{participant}_reference_labels_{channel}.csv"
        )
        reference_data = pd.read_csv(file_path)
        reference_data = reference_data.set_index(["heartbeat_id", "channel", "label"])
        # absolute and relative sample are the same in the example data
        reference_data = reference_data.assign(sample_absolute=reference_data["sample_relative"])
        return reference_data

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        """Return the reference heartbeats.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference heartbeats as a pandas DataFrame

        """
        return self._load_reference_heartbeats()

    def _load_reference_heartbeats(self) -> pd.DataFrame:
        reference_ecg = self.reference_labels_ecg
        reference_heartbeats = reference_ecg.reindex(["heartbeat"], level="channel")
        reference_heartbeats = compute_reference_heartbeats(reference_heartbeats)
        return reference_heartbeats



import ast

import pandas as pd

from pepbench.utils._types import path_t

__all__ = ["compute_reference_heartbeats", "load_labeling_borders"]


def load_labeling_borders(file_path: path_t) -> pd.DataFrame:
    """Load the labeling borders from a csv file.

    Parameters
    ----------
    file_path : :class:`pathlib.Path` or str
        The path to the csv file.

    Returns
    -------
    :class:`pandas.DataFrame`
        The labeling borders.

    """
    data = pd.read_csv(file_path)
    data = data.assign(description=data["description"].apply(lambda s: ast.literal_eval(s)))

    data = data.set_index("timestamp").sort_index()
    return data


def compute_reference_heartbeats(heartbeats: pd.DataFrame) -> pd.DataFrame:
    """Reformat the heartbeats DataFrame.

    Parameters
    ----------
    heartbeats : :class:`pandas.DataFrame`
        DataFrame containing the heartbeats.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame containing the reformatted heartbeats.

    """
    heartbeats = heartbeats.droplevel("channel")["sample_relative"].unstack("label")
    heartbeats.columns = [f"{col}_sample" for col in heartbeats.columns]
    return heartbeats



from collections.abc import Sequence

import numpy as np
import pandas as pd
import pingouin as pg

from pepbench.utils._types import str_t

__all__ = [
    "add_unique_id_to_results_dataframe",
    "compute_improvement_outlier_correction",
    "compute_improvement_pipeline",
    "compute_pep_performance_metrics",
    "correlation_reference_pep_heart_rate",
    "describe_pep_values",
    "get_data_for_algo",
    "get_error_by_group",
    "get_pep_for_algo",
    "get_reference_data",
    "get_reference_pep",
    "rr_interval_to_heart_rate",
]


_algo_levels = ["q_peak_algorithm", "b_point_algorithm", "outlier_correction_algorithm"]

_pep_error_metric_map = {
    "absolute_error_per_sample_ms": "Mean Absolute Error [ms]",
    "error_per_sample_ms": "Mean Error [ms]",
    "absolute_relative_error_per_sample_percent": "Mean Absolute Relative Error [%]",
}
_pep_number_map = {
    "num_pep_valid": "Valid PEPs",
    "num_pep_invalid": "Invalid PEPs",
    "num_pep_total": "Total PEPs",
}


def get_reference_data(results_per_sample: pd.DataFrame) -> pd.DataFrame:
    """Extract the reference data from the *results-per-sample* dataframe.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.

    Returns
    -------
    :class:`pandas.DataFrame`
        The reference data.

    """
    reference_pep = results_per_sample.xs("reference", level=-1, axis=1)
    reference_pep = reference_pep.groupby(_algo_levels)
    reference_pep = reference_pep.get_group(next(iter(reference_pep.groups))).droplevel(_algo_levels)

    return reference_pep


def get_reference_pep(results_per_sample: pd.DataFrame) -> pd.DataFrame:
    """Extract the reference PEP values from the *results-per-sample* dataframe.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.

    Returns
    -------
    :class:`pandas.DataFrame`
        The reference PEP values.

    """
    return get_reference_data(results_per_sample)[["pep_ms"]]


def get_data_for_algo(results_per_sample: pd.DataFrame, algo_combi: str_t) -> pd.DataFrame:
    """Extract the data for a specific algorithm combination from the *results-per-sample* dataframe.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.
    algo_combi : str or tuple of str
        The algorithm combination for which the data should be extracted.

    Returns
    -------
    :class:`pandas.DataFrame`
        The data for the specified algorithm combination.

    """
    algo_levels = [s for s in results_per_sample.index.names if s in _algo_levels]
    if isinstance(algo_combi, str):
        algo_combi = (algo_combi,)
    data = results_per_sample.xs(tuple(algo_combi), level=algo_levels)
    return data


def get_pep_for_algo(results_per_sample: pd.DataFrame, algo_combi: Sequence[str]) -> pd.DataFrame:
    """Extract the PEP values for a specific algorithm combination from the *results-per-sample* dataframe.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.
    algo_combi : str or tuple of str
        The algorithm combination for which the PEP values should be extracted.

    Returns
    -------
    :class:`pandas.DataFrame`
        The PEP values for the specified algorithm combination.

    """
    pep = get_data_for_algo(results_per_sample, algo_combi)
    pep = pep[[("pep_ms", "estimated")]].droplevel(level=-1, axis=1)

    return pep


def describe_pep_values(
    data: pd.DataFrame, group_cols: str_t | None = None, metrics: Sequence[str] | None = None
) -> pd.DataFrame:
    """Compute the descriptive statistics for the PEP values using the :meth:`pandas.DataFrame.describe` method.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The PEP values.
    group_cols : str or list of str, optional
        The column(s) to group the data by, if any. Default: "phase".
    metrics : list of str, optional
        List of metrics to display from the descriptive statistics. Default: ["mean", "std", "min", "max"].

    """
    if group_cols is None:
        group_cols = ["phase"]
    if metrics is None:
        metrics = ["mean", "std", "min", "max"]

    return data.groupby(group_cols).describe().reindex(metrics, level=-1, axis=1)


def compute_pep_performance_metrics(
    results_per_sample: pd.DataFrame,
    *,
    num_heartbeats: pd.DataFrame | None = None,
    metrics: Sequence[str] | None = None,
    sortby: str_t | None = ("absolute_error_per_sample_ms", "mean"),
    ascending: bool | None = True,
) -> pd.DataFrame:
    """Compute the performance metrics for the PEP values.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.
    num_heartbeats : :class:`pandas.DataFrame`, optional
        Dataframe containing the number of heartbeats (to include in the output). Default: None.
    metrics : list of str, optional
        List of metrics to compute. Default: ["mean", "std"].
    sortby : str, optional
        The column to sort the results by. Default: ("absolute_error_per_sample_ms", "mean").
    ascending : bool, optional
        Whether to sort the results in ascending order. Default: True.

    """
    if metrics is None:
        metrics = ["mean", "std"]
    results_per_sample = results_per_sample.copy()
    algo_levels = [s for s in results_per_sample.index.names if s in _algo_levels]
    results_per_sample = results_per_sample[_pep_error_metric_map.keys()].droplevel(level=-1, axis=1)
    results_per_sample = results_per_sample.groupby(algo_levels)
    results_per_sample = results_per_sample.agg(metrics)

    num_heartbeats = num_heartbeats.unstack().swaplevel(axis=1)
    results_per_sample = results_per_sample.join(num_heartbeats)

    if sortby is not None:
        results_per_sample = results_per_sample.sort_values(sortby, ascending=ascending)

    rename_map = _pep_error_metric_map.copy()
    rename_map.update(_pep_number_map)
    results_per_sample = results_per_sample.rename(rename_map, level=0, axis=1)
    results_per_sample = results_per_sample.reindex(rename_map.values(), level=0, axis=1)

    return results_per_sample


def get_performance_metric(results_per_sample: pd.DataFrame, metric: str) -> pd.DataFrame:
    """Extract a specific performance metric from the *results-per-sample* dataframe.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.
    metric : str

    Returns
    -------
    :class:`pandas.DataFrame`
        The extracted performance metric.

    """
    return results_per_sample[[metric]].droplevel(level=-1, axis=1)


def rr_interval_to_heart_rate(data: pd.DataFrame) -> pd.DataFrame:
    """Convert RR intervals in milliseconds to heart rate in beats per minute.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The data containing the RR intervals in milliseconds. The column name must be "rr_interval_ms".

    Returns
    -------
    :class:`pandas.DataFrame`
        The data with the heart rate in beats per minute

    """
    heart_rate_bpm = 60 * 1000 / data[["rr_interval_ms"]]
    heart_rate_bpm = heart_rate_bpm.rename(columns={"rr_interval_ms": "heart_rate_bpm"})
    return data.join(heart_rate_bpm)


def correlation_reference_pep_heart_rate(data: pd.DataFrame, groupby: str | None = None) -> dict[str, pd.DataFrame]:
    """Compute the correlation between the reference PEP values and the heart rate.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The data containing the reference PEP values and the heart rate.
    groupby : str, optional
        The column to group the data by. If None, no grouping is applied. Default: None.

    Returns
    -------
    dict
        A dictionary containing the linear regression model and the correlation coefficient.

    """
    data = get_reference_data(data)

    # compute a linear regression model
    if groupby is None:
        linreg = pg.linear_regression(X=data["heart_rate_bpm"], y=data["pep_ms"], remove_na=True)
        corr = pg.corr(data["heart_rate_bpm"], data["pep_ms"], method="pearson")
    else:
        linreg = data.groupby(groupby).apply(
            lambda df: pg.linear_regression(X=df["heart_rate_bpm"], y=df["pep_ms"], remove_na=True)
        )
        corr = data.groupby(groupby).apply(lambda df: pg.corr(df["heart_rate_bpm"], df["pep_ms"], method="pearson"))

    return {"linear_regression": linreg, "correlation": corr}


def get_error_by_group(
    results_per_sample: pd.DataFrame, error_metric: str = "absolute_error_per_sample_ms", grouper: str_t = "participant"
) -> pd.DataFrame:
    """Compute mean and standard deviation of the error metric by group.

    Parameters
    ----------
    results_per_sample : :class:`pandas.DataFrame`
        The results-per-sample dataframe.
    error_metric : str, optional
        The error metric to extract. Default: "absolute_error_per_sample_ms".
    grouper : str or list of str, optional
        The column(s) to group the data by. Default: "participant".

    Returns
    -------
    :class:`pandas.DataFrame`
        The error metric aggregated by group.

    """
    algo_levels = [s for s in results_per_sample.index.names if s in _algo_levels]
    if isinstance(grouper, str):
        grouper = [grouper]

    error = results_per_sample[[error_metric]].groupby([*algo_levels, *grouper]).agg(["mean", "std"])
    error = (
        error.droplevel(1, axis=1)
        .unstack(algo_levels)
        .reorder_levels([0, *range(2, 2 + len(algo_levels)), 1], axis=1)
        .sort_index(axis=1)
    )
    error.columns = error.columns.set_names("metric", level=0)
    return error


def add_unique_id_to_results_dataframe(data: pd.DataFrame, algo_levels: Sequence[str] | None = None) -> pd.DataFrame:
    """Add a unique ID to the results dataframe.

    The unique ID is created by concatenating the values of the specified algorithm levels and the heartbeat IDs.
    This is then added as a new index level named "id_concat".

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The results dataframe.
    algo_levels : list of str, optional
        The algorithm levels to use for the unique ID. If None, the default algorithm levels
        (["q_peak_algorithm", "b_point_algorithm", "outlier_correction_algorithm"]) are used.

    Returns
    -------
    :class:`pandas.DataFrame`
        The results dataframe with the unique IDs added as new index level.

    """
    data = data.copy()
    data = data.droplevel(axis=1, level=-1).rename(index=str)
    if algo_levels is None:
        algo_levels = _algo_levels
    if isinstance(algo_levels, str):
        algo_levels = [algo_levels]

    algo_levels = [s for s in data.index.names if s in algo_levels]
    data = data.reset_index(level=algo_levels)
    id_concat = pd.Index(["_".join(i) for i in data.index], name="id_concat")
    data = data.assign(id_concat=id_concat)
    data = data.set_index([*algo_levels, "id_concat"])
    return data


def compute_improvement_outlier_correction(data: pd.DataFrame, outlier_algos: Sequence[str]) -> pd.DataFrame:
    """Compute the percentage of samples which improved, deteriorated, or remained unchanged after outlier correction.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The data containing the PEP values before and after outlier correction.
    outlier_algos : list of str
        The outlier correction algorithms to consider.

    Returns
    -------
    :class:`pandas.DataFrame`
        The percentage of samples which improved, deteriorated, or remained unchanged after outlier correction.

    """
    data = data.copy()
    if "outlier_correction_algorithm" not in data.columns.names:
        data = data.unstack("outlier_correction_algorithm")
    data = data.reindex(outlier_algos, axis=1, level="outlier_correction_algorithm")
    data = data.diff(axis=1).dropna(how="all", axis=1)
    data = np.sign(data)
    data.columns = ["improvement_percent"]
    # negative values = improvement through outlier correction
    # count the number of positive and negative values
    improvement_percent = data.value_counts(normalize=True) * 100
    improvement_percent = improvement_percent.rename({-1: "improvement", 1: "deterioration", 0: "no change"})
    improvement_percent = improvement_percent.to_frame().reindex(["improvement", "no change", "deterioration"], level=0)
    improvement_percent.columns = ["improvement_percent"]
    improvement_percent.index.names = [None]
    improvement_percent = improvement_percent.T
    return improvement_percent


def compute_improvement_pipeline(data: pd.DataFrame, pipelines: Sequence[str]) -> pd.DataFrame:
    """Compute the percentage of samples which showed sign changes in the error metric between two pipelines.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        The data containing the PEP extraction results from different pipelines.
    pipelines : list of str
        The pipelines to compare.

    Returns
    -------
    :class:`pandas.DataFrame`
        Overview of the percentage of samples which showed a change in the sign of the error metric
        (i.e., either positive to negative, vice versa, or no change) between two pipelines.

    """
    data = data.copy()
    pipelines = ["_".join(i) for i in pipelines]

    data = data.unstack("pipeline").reindex(pipelines, level="pipeline", axis=1)
    # compute the percentage of sample which have a positive value in the first column
    # and a negative value in the second column
    data = data.assign(change_pos_neg=(data.iloc[:, 0] > 0) & (data.iloc[:, 1] < 0))
    data = data.assign(change_pos_pos=(data.iloc[:, 0] > 0) & (data.iloc[:, 1] > 0))
    data = data.assign(change_neg_pos=(data.iloc[:, 0] < 0) & (data.iloc[:, 1] > 0))
    data = data.assign(change_neg_neg=(data.iloc[:, 0] < 0) & (data.iloc[:, 1] < 0))
    data = data.assign(change_diff=data["change_pos_neg"] | data["change_neg_pos"])
    data = data.assign(change_same=data["change_pos_pos"] | data["change_neg_neg"])

    data = data.filter(like="change", axis=1)
    data = data.apply(pd.Series.value_counts, normalize=True) * 100

    return data



"""Datasets for beat-to-beat PEP extraction.

Currently, two following datasets are available in `pepbench` (see below). Both datasets provide a unified interface
to the underlying data and enable easy use with the PEP extraction algorithms and pipelines.

The PEP extraction pipelines require the datasets to provide the following attributes:

    - `ecg`: Access to the ECG signal data
    - `icg`: Access to the ICG signal data
    - `sampling_rate_ecg`: Sampling rate of the ECG signal
    - `sampling_rate_icg`: Sampling rate of the ICG signal

The provided dataset classes are implemented as subclasses of :class:`pepbench.datasets.BasePepDataset`.
This interface provides these necessary attributes and methods to extract PEP data from the underlying signals.

If the datasets should be used with labeled reference data, i.e., for evaluating the PEP extraction algorithms,
the datasets should also provide the following attributes:

    - `reference_pep`: Reference PEP data, which should have the following columns:
        - `heartbeat_start_sample`: Index of the heartbeat start sample, relative to the beginning of the signal
        - `heartbeat_end_sample`: Index of the heartbeat end sample, relative to the beginning of the signal
        - `q_peak_sample`: Index of the Q-peak sample, relative to the beginning of the signal
        - `b_point_sample`: Index of the B-point sample, relative to the beginning of the signal
    - `reference_heartbeats`: Reference heartbeats, which should have the following columns:
        - `heartbeat_start_sample`: Index of the heartbeat start sample, relative to the beginning of the signal
        - `heartbeat_end_sample`: Index of the heartbeat end sample, relative to the beginning of the signal
    - `reference_labels_ecg`: Reference labels for the ECG data
    - `reference_labels_icg`: Reference labels for the ICG data


"""

from pepbench.datasets._base_pep_extraction_dataset import (
    BasePepDataset,
    BasePepDatasetWithAnnotations,
    MetadataMixin,
    PepLabelMixin,
)
from pepbench.datasets.empkins import EmpkinsDataset
from pepbench.datasets.guardian import GuardianDataset
from pepbench.datasets.rebeat_icg import ReBeatIcgDataset
from pepbench.datasets.time_window_icg import TimeWindowIcgDataset

__all__ = [
    "BasePepDataset",
    "BasePepDatasetWithAnnotations",
    "EmpkinsDataset",
    "GuardianDataset",
    "MetadataMixin",
    "PepLabelMixin",
    "ReBeatIcgDataset",
    "TimeWindowIcgDataset",
]



import inspect
from collections.abc import Sequence
from typing import Any

import numpy as np
import pandas as pd
from fau_colors.v2021 import cmaps
from matplotlib import pyplot as plt
from matplotlib.patches import FancyBboxPatch
from matplotlib.transforms import Bbox

from pepbench.datasets import BasePepDataset, BasePepDatasetWithAnnotations
from pepbench.heartbeat_matching import match_heartbeat_lists


def _get_fig_ax(kwargs: dict[str, Any]) -> tuple[plt.Figure, plt.Axes]:
    ax = kwargs.pop("ax", None)
    # filter out the kwargs that are not needed for plt.subplots
    kwargs_subplot = {
        key: value
        for key, value in kwargs.items()
        if key
        in list(inspect.signature(plt.subplots).parameters.keys())
        + list(inspect.signature(plt.figure).parameters.keys())
    }
    # remove the kwargs that are used for plt.subplots
    for key in kwargs_subplot:
        kwargs.pop(key)

    if ax is not None:
        fig = ax.get_figure()
    else:
        fig, ax = plt.subplots(**kwargs_subplot)

    return fig, ax


def _get_fig_axs(kwargs: dict[str, Any]) -> tuple[plt.Figure, Sequence[plt.Axes]]:
    axs: plt.Axes | None = kwargs.pop("axs", None)
    # filter out the kwargs that are not needed for plt.subplots
    kwargs_subplot = {
        key: value
        for key, value in kwargs.items()
        if key
        in list(inspect.signature(plt.subplots).parameters.keys())
        + list(inspect.signature(plt.figure).parameters.keys())
    }

    if axs is not None:
        fig = axs[0].get_figure()
    else:
        fig, axs = plt.subplots(**kwargs_subplot)
    return fig, axs


def _add_ecg_r_peaks(
    ecg_data: pd.DataFrame,
    r_peaks: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    label = kwargs.get("r_peak_label", "R-Peaks")
    color = kwargs.get("r_peak_color", cmaps.wiso[0])
    marker = kwargs.get("r_peak_marker", "o")
    linestyle = kwargs.get("r_peak_linestyle", "-")
    linewidth = kwargs.get("r_peak_linewidth")
    alpha = kwargs.get("r_peak_alpha", 0.7)
    plot_marker = kwargs.get("r_peak_plot_marker", True)
    plot_vline = kwargs.get("r_peak_plot_vline", True)

    r_peaks = r_peaks.astype(int)

    if plot_marker:
        _base_add_scatter(
            x=ecg_data.index[r_peaks], y=ecg_data.iloc[r_peaks], color=color, label=label, marker=marker, ax=ax
        )
    if plot_vline:
        _base_add_vlines(
            x=ecg_data.index[r_peaks],
            color=color,
            alpha=alpha,
            linestyle=linestyle,
            linewidth=linewidth,
            label=label,
            ax=ax,
        )


def _add_ecg_r_peak_artefacts(
    ecg_data: pd.DataFrame,
    r_peaks: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    label = kwargs.get("r_peak_artefact_label", "R-Peak Artefacts")
    color = kwargs.get("r_peak_artefact_color", cmaps.wiso_dark[0])
    marker = kwargs.get("r_peak_artefact_marker", "X")
    linestyle = kwargs.get("r_peak_artefact_linestyle", "-")

    kwargs = kwargs.copy()
    kwargs.update(r_peak_label=label, r_peak_color=color, r_peak_marker=marker, r_peak_linestyle=linestyle)

    _add_ecg_r_peaks(
        ecg_data,
        r_peaks,
        ax,
        **kwargs,
    )


def _add_ecg_q_peaks(
    ecg_data: pd.DataFrame,
    q_peaks: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    label = kwargs.get("q_peak_label", "Q-Peaks")
    color = kwargs.get("q_peak_color", cmaps.med[0])
    marker = kwargs.get("q_peak_marker", "o")
    linestyle = kwargs.get("q_peak_linestyle", "-")
    linewidth = kwargs.get("q_peak_linewidth")
    alpha = kwargs.get("q_peak_alpha", 0.7)

    q_peaks = q_peaks.astype(int)
    if len(q_peaks) == 0:
        return

    _base_add_scatter(
        x=ecg_data.index[q_peaks], y=ecg_data.iloc[q_peaks], color=color, label=label, marker=marker, ax=ax
    )
    _base_add_vlines(
        x=ecg_data.index[q_peaks],
        color=color,
        alpha=alpha,
        label=label,
        linestyle=linestyle,
        linewidth=linewidth,
        ax=ax,
    )


def _add_ecg_q_peak_artefacts(
    ecg_data: pd.DataFrame,
    q_peak_artefacts: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    label = kwargs.get("q_peak_artefact_label", "Q-Peak Artefacts")
    color = kwargs.get("q_peak_artefact_color", cmaps.med_dark[0])
    marker = kwargs.get("q_peak_artefact_marker", "X")
    linestyle = kwargs.get("q_peak_artefact_linestyle", "-")

    kwargs = kwargs.copy()
    kwargs.update(q_peak_label=label, q_peak_color=color, q_peak_marker=marker, q_peak_linestyle=linestyle)

    _add_ecg_q_peaks(
        ecg_data,
        q_peak_artefacts,
        ax,
        **kwargs,
    )


def _add_icg_b_points(
    icg_data: pd.DataFrame,
    b_points: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    color = kwargs.get("b_point_color", cmaps.phil[0])
    b_point_label = kwargs.get("b_point_label", "B-Points")
    marker = kwargs.get("b_point_marker", "o")
    linestyle = kwargs.get("b_point_linestyle", "-")
    alpha = kwargs.get("b_point_alpha", 0.7)

    b_points = b_points.astype(int)
    b_points = np.atleast_1d(b_points)

    if len(b_points) == 0:
        return

    if kwargs.get("b_point_plot_marker", True):
        _base_add_scatter(
            x=icg_data.index[b_points],
            y=icg_data.iloc[b_points],
            color=color,
            label=b_point_label,
            marker=marker,
            ax=ax,
        )
    if kwargs.get("b_point_plot_vlines", True):
        _base_add_vlines(
            x=icg_data.index[b_points], color=color, alpha=alpha, label=b_point_label, linestyle=linestyle, ax=ax
        )


def _add_icg_b_point_artefacts(
    icg_data: pd.DataFrame,
    b_points: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    color = kwargs.get("b_point_artefact_color", cmaps.phil_dark[0])
    label = kwargs.get("b_point_artefact_label", "B-Point Artefacts")
    marker = kwargs.get("b_point_artefact_marker", "X")
    linestyle = kwargs.get("b_point_artefact_linestyle", "-")

    kwargs = kwargs.copy()
    kwargs.update(b_point_color=color, b_point_label=label, b_point_marker=marker, b_point_linestyle=linestyle)

    _add_icg_b_points(icg_data, b_points, ax, **kwargs)


def _add_icg_c_points(
    icg_data: pd.DataFrame,
    c_points: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    color = kwargs.get("c_point_color", cmaps.wiso_light[0])
    c_point_label = kwargs.get("c_point_label", "C Points")
    marker = kwargs.get("c_point_marker", "X")
    linestyle = kwargs.get("c_point_linestyle", "--")
    linewidth = kwargs.get("c_point_linewidth")
    alpha = kwargs.get("c_point_alpha", 0.7)
    plot_marker = kwargs.get("c_point_plot_marker", True)
    plot_vline = kwargs.get("c_point_plot_vline", True)

    c_points = c_points.astype(int)

    if plot_marker:
        _base_add_scatter(
            x=icg_data.index[c_points],
            y=icg_data.iloc[c_points],
            color=color,
            label=c_point_label,
            marker=marker,
            ax=ax,
        )
    if plot_vline:
        _base_add_vlines(
            x=icg_data.index[c_points],
            color=color,
            alpha=alpha,
            label=c_point_label,
            linestyle=linestyle,
            linewidth=linewidth,
            ax=ax,
        )


def _base_add_vlines(
    x: pd.Series,
    color: str,
    alpha: float,
    label: str,
    linestyle: str,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    ax.vlines(
        x=x,
        ymin=0,
        ymax=1,
        color=color,
        transform=ax.get_xaxis_transform(),
        zorder=3,
        alpha=alpha,
        ls=linestyle,
        label=label,
        **kwargs,
    )


def _base_add_scatter(
    x: pd.Series,
    y: pd.Series,
    color: str,
    label: str,
    marker: str,
    ax: plt.Axes,
) -> None:
    ax.scatter(
        x=x,
        y=y,
        color=color,
        label=label,
        zorder=3,
        marker=marker,
    )


def _add_pep_from_reference(
    ecg_data: pd.DataFrame,
    icg_data: pd.DataFrame,
    labels: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    color = kwargs.get("pep_color", cmaps.nat[0])
    color_artefact = kwargs.get("pep_artefact_color", cmaps.wiso[0])
    edgecolor = kwargs.get("pep_edgecolor", color)
    facecolor = kwargs.get("pep_facecolor", color)
    edgecolor_artefact = kwargs.get("pep_artefact_edgecolor", color_artefact)
    facecolor_artefact = kwargs.get("pep_artefact_facecolor", color_artefact)
    pep_label = kwargs.get("pep_label", "PEP")

    hatch = kwargs.get("pep_hatch")
    if hatch is not None:
        facecolor = "none"
        facecolor_artefact = "none"

    for _i, row in labels.iterrows():
        start = ecg_data.index[row[("ecg", "sample_relative")]]
        end = icg_data.index[row[("icg", "sample_relative")]]

        if row[("ecg", "label")] == "Artefact" or row[("icg", "label")] == "Artefact":
            ax.axvspan(
                start,
                end,
                edgecolor=edgecolor_artefact,
                facecolor=facecolor_artefact,
                hatch=hatch,
                alpha=0.3,
                zorder=0,
                label=f"{pep_label} Artefact",
            )
        else:
            ax.axvspan(
                start,
                end,
                edgecolor=edgecolor,
                facecolor=facecolor,
                hatch=hatch,
                alpha=0.3,
                zorder=0,
                label=pep_label,
            )


def _add_pep_from_results(
    ecg_data: pd.DataFrame,
    icg_data: pd.DataFrame,
    labels: pd.DataFrame,
    ax: plt.Axes,
    **kwargs: Any,
) -> None:
    color = kwargs.get("pep_color", cmaps.nat[0])
    edgecolor = kwargs.get("pep_edgecolor", color)
    facecolor = kwargs.get("pep_facecolor", color)
    label = kwargs.get("pep_label", "PEP")

    hatch = kwargs.get("pep_hatch")
    if hatch is not None:
        facecolor = "none"

    for _i, row in labels.iterrows():
        start = ecg_data.index[row["ecg"]]
        end = icg_data.index[row["icg"]]
        ax.axvspan(
            start,
            end,
            facecolor=facecolor,
            edgecolor=edgecolor,
            alpha=0.3,
            zorder=0,
            hatch=hatch,
            label=label,
        )


def _add_heartbeat_borders(heartbeats: pd.DataFrame, ax: plt.Axes, **kwargs: Any) -> None:
    color = kwargs.get("heartbeat_border_color", cmaps.tech[2])
    ax.vlines(
        x=heartbeats,
        ymin=0,
        ymax=1,
        zorder=0,
        colors=color,
        transform=ax.get_xaxis_transform(),
        label="Heartbeat Borders",
        ls="--",
    )


def _handle_legend_one_axis(fig: plt.Figure, ax: plt.Axes, **kwargs: Any) -> None:
    legend_max_cols = kwargs.get("legend_max_cols", 5)
    legend_outside = kwargs.get("legend_outside", False)
    legend_orientation = kwargs.get("legend_orientation", "vertical")
    legend_loc = kwargs.get("legend_loc")

    handles, labels = ax.get_legend_handles_labels()
    handles, labels = _remove_duplicate_legend_entries(handles, labels)
    ncols = min(len(handles), legend_max_cols) if legend_orientation == "horizontal" else 1

    if len(fig.legends) > 0:
        fig.legends[0].remove()

    if legend_outside:
        ax.legend().remove()
        fig.legend(handles=handles, labels=labels, ncols=ncols, loc=legend_loc)
    else:
        ax.legend(handles=handles, labels=labels, loc=legend_loc, ncols=ncols)


def _handle_legend_two_axes(
    fig: plt.Figure,
    axs: Sequence[plt.Axes],
    **kwargs: Any,
) -> None:
    legend_max_cols = kwargs.get("legend_max_cols", 5)
    legend_outside = kwargs.get("legend_outside", False)
    legend_orientation = kwargs.get("legend_orientation", "vertical")
    legend_loc = kwargs.get("legend_loc")

    if len(fig.legends) > 0:
        fig.legends[0].remove()
    if legend_outside:
        handles, labels = axs[0].get_legend_handles_labels()
        for _i, ax in enumerate(axs[1:]):
            handles += ax.get_legend_handles_labels()[0]
            labels += ax.get_legend_handles_labels()[1]
        for ax in axs:
            ax.legend().remove()

        ncols = min(len(handles), legend_max_cols) if legend_orientation == "horizontal" else 1
        handles, labels = _remove_duplicate_legend_entries(handles, labels)

        fig.legend(handles, labels, ncols=ncols, loc=legend_loc)
    else:
        for ax in axs:
            _handle_legend_one_axis(fig, ax, **kwargs)


def _remove_duplicate_legend_entries(handles: Sequence[plt.Artist], labels: Sequence[str]) -> tuple[list, list]:
    unique_labels = []
    unique_handles = []
    for handle, label in zip(handles, labels, strict=False):
        if label not in unique_labels:
            unique_labels.append(label)
            unique_handles.append(handle)
    return unique_handles, unique_labels


def _sanitize_heartbeat_subset(heartbeat_subsample: Sequence[int] | None = None) -> Sequence[int] | None:
    if heartbeat_subsample is None:
        return None
    if len(heartbeat_subsample) == 1:
        return heartbeat_subsample
    # if heartbeat_subsample is a tuple of two integers, assume it's a range
    if isinstance(heartbeat_subsample, tuple) and len(heartbeat_subsample) == 2:
        return list(range(heartbeat_subsample[0], heartbeat_subsample[1] + 1))
    # if it's a sequence but not incremented by 1, it's an error
    if np.ediff1d(heartbeat_subsample).max() != 1:
        raise ValueError("Heartbeat subsample must be a range or list of indices.")
    return heartbeat_subsample


def _get_data(
    datapoint: BasePepDataset, *, normalize_time: bool, heartbeat_subset: Sequence[int] | None
) -> tuple[pd.DataFrame, pd.DataFrame]:
    ecg_data = datapoint.ecg
    icg_data = datapoint.icg

    if isinstance(ecg_data.index, pd.TimedeltaIndex):
        ecg_data.index = ecg_data.index.total_seconds()
    if isinstance(icg_data.index, pd.TimedeltaIndex):
        icg_data.index = icg_data.index.total_seconds()

    heartbeat_subset = _sanitize_heartbeat_subset(heartbeat_subset)

    if heartbeat_subset is None:
        heartbeat_borders = datapoint.heartbeats
        heartbeat_subset = list(heartbeat_borders.index)
    else:
        heartbeat_borders = datapoint.heartbeats

    start = heartbeat_borders["start_sample"].iloc[heartbeat_subset[0]]
    end = heartbeat_borders["end_sample"].iloc[heartbeat_subset[-1]]

    ecg_data = ecg_data.iloc[start:end]
    icg_data = icg_data.iloc[start:end]

    if normalize_time:
        if isinstance(ecg_data.index, pd.TimedeltaIndex | pd.DatetimeIndex):
            ecg_data.index = (ecg_data.index - ecg_data.index[0]).total_seconds()
        else:
            ecg_data.index -= ecg_data.index[0]
        if isinstance(icg_data.index, pd.TimedeltaIndex | pd.DatetimeIndex):
            icg_data.index = (icg_data.index - icg_data.index[0]).total_seconds()
        else:
            icg_data.index -= icg_data.index[0]

    return ecg_data, icg_data


def _get_reference_labels(
    datapoint: BasePepDatasetWithAnnotations,
    heartbeat_subset: Sequence[int] | None = None,
) -> dict[str, pd.DataFrame]:
    heartbeats = datapoint.heartbeats
    heartbeats = heartbeats[["start_sample", "end_sample"]]
    reference_heartbeats = datapoint.reference_heartbeats
    reference_labels_ecg = datapoint.reference_labels_ecg
    reference_labels_icg = datapoint.reference_labels_icg

    q_peaks = reference_labels_ecg.reindex(["ECG"], level="channel")["sample_relative"]
    q_peak_artefacts = reference_labels_ecg.reindex(["Artefact"], level="channel")["sample_relative"]
    b_points = reference_labels_icg.reindex(["ICG"], level="channel")["sample_relative"]
    b_point_artefacts = reference_labels_icg.reindex(["Artefact"], level="channel")["sample_relative"]

    heartbeat_subset = _sanitize_heartbeat_subset(heartbeat_subset)

    # match the heartbeats to the reference labels
    heartbeat_matching = match_heartbeat_lists(
        heartbeats_reference=reference_heartbeats,
        heartbeats_extracted=heartbeats,
        tolerance_ms=100,
        sampling_rate_hz=datapoint.sampling_rate_ecg,
    )

    tp_matches = heartbeat_matching.query("match_type == 'tp'")

    reference_heartbeats = reference_heartbeats.loc[tp_matches["heartbeat_id_reference"]]
    heartbeats = heartbeats.loc[tp_matches["heartbeat_id"]]

    if heartbeat_subset is not None:
        reference_heartbeats = reference_heartbeats.iloc[heartbeat_subset]
        heartbeats = heartbeats.iloc[heartbeat_subset]
        q_peaks = q_peaks.reindex(reference_heartbeats.index, level="heartbeat_id").dropna()
        q_peak_artefacts = q_peak_artefacts.reindex(reference_heartbeats.index, level="heartbeat_id").dropna()
        b_points = b_points.reindex(reference_heartbeats.index, level="heartbeat_id").dropna()
        b_point_artefacts = b_point_artefacts.reindex(reference_heartbeats.index, level="heartbeat_id").dropna()

    return_dict = {
        "heartbeats": heartbeats,
        "q_peaks": q_peaks,
        "q_peak_artefacts": q_peak_artefacts,
        "b_points": b_points,
        "b_point_artefacts": b_point_artefacts,
    }

    # compute the start_sample of the first heartbeat in the subset to get relative sample indices, corrected by the
    # offset between the reference heartbeat borders and the extracted heartbeat borders
    start_sample = reference_heartbeats["start_sample"].iloc[0] - (
        reference_heartbeats["start_sample"].iloc[0] - heartbeats["start_sample"].iloc[0]
    )

    # subtract start_sample to get relative sample indices
    return {key: value - start_sample for key, value in return_dict.items()}


def _get_labels_from_challenge_results(
    pep_results_per_sample: pd.DataFrame, heartbeat_subsample: Sequence[int]
) -> dict[str, pd.DataFrame]:
    heartbeat_subsample = _sanitize_heartbeat_subset(heartbeat_subsample)
    pep_results_per_sample = pep_results_per_sample.reindex(heartbeat_subsample, level="heartbeat_id")

    heartbeats_start = pep_results_per_sample[("heartbeat_start_sample", "estimated")]
    heartbeats_end = pep_results_per_sample[("heartbeat_end_sample", "estimated")]

    start_index = heartbeats_start.iloc[0]

    q_peak_labels_reference = pep_results_per_sample[("q_peak_sample", "reference")]
    q_peak_labels_estimated = pep_results_per_sample[("q_peak_sample", "estimated")]
    b_point_labels_reference = pep_results_per_sample[("b_point_sample", "reference")]
    b_point_labels_estimated = pep_results_per_sample[("b_point_sample", "estimated")]

    return_dict = {
        "heartbeats_start": heartbeats_start,
        "heartbeats_end": heartbeats_end,
        "q_peak_labels_reference": q_peak_labels_reference,
        "q_peak_labels_estimated": q_peak_labels_estimated,
        "b_point_labels_reference": b_point_labels_reference,
        "b_point_labels_estimated": b_point_labels_estimated,
    }
    return {key: (val.dropna() - start_index).astype(int) for key, val in return_dict.items()}


def _get_rect(kwargs: dict[str, Any]) -> tuple[float, ...]:
    rect = kwargs.pop("rect", None)
    legend_outside = kwargs.get("legend_outside", False)
    legend_orientation = kwargs.get("legend_orientation", "vertical")

    if rect is not None:
        return rect
    if legend_outside:
        return (0, 0, 0.80, 1.0) if legend_orientation == "vertical" else (0, 0, 1, 0.90)
    return (0, 0, 1, 1)


def _get_legend_loc(kwargs: dict[str, Any]) -> str:
    legend_loc = kwargs.get("legend_loc")
    legend_outside = kwargs.get("legend_outside", False)
    legend_orientation = kwargs.get("legend_orientation", "vertical")

    if legend_loc is not None:
        return legend_loc
    if legend_outside:
        return "upper right" if legend_orientation == "vertical" else "upper center"
    return "upper right" if legend_orientation == "vertical" else "lower center"


def _get_annotation_bbox() -> dict[str, Any]:
    return {
        "fc": (1, 1, 1, plt.rcParams["legend.framealpha"]),
        "ec": plt.rcParams["legend.edgecolor"],
        "boxstyle": "round",
    }


def _get_annotation_bbox_no_edge() -> dict[str, Any]:
    return {
        "fc": (1, 1, 1, plt.rcParams["legend.framealpha"]),
        "ec": "none",
        "boxstyle": "round",
    }


def _get_bbox_coords(
    artist: plt.Artist,
    ax: plt.Axes,
) -> np.ndarray:
    return artist.get_window_extent().transformed(ax.transData.inverted()).get_points()


def add_fancy_patch_around(
    ax: plt.Axes,
    bb: Bbox,
    **kwargs: dict[str, Any],
) -> FancyBboxPatch:
    kwargs.setdefault("fc", (1, 1, 1, plt.rcParams["legend.framealpha"]))
    kwargs.setdefault("ec", "none")
    kwargs.setdefault("boxstyle", "round,pad=5")
    fancy = FancyBboxPatch(bb.p0, bb.width, bb.height, **kwargs)
    ax.add_patch(fancy)
    return fancy


def _get_heartbeats(
    datapoint: BasePepDatasetWithAnnotations, heartbeat_subset: Sequence[int] | None = None, normalize: bool = True
) -> pd.DataFrame:
    heartbeats = datapoint.heartbeats.drop(columns="start_time")
    if heartbeat_subset is not None:
        heartbeats = heartbeats.loc[heartbeat_subset][["start_sample", "end_sample", "r_peak_sample"]]
        heartbeats = (heartbeats - heartbeats.iloc[0]["start_sample"]).astype(int)

    if normalize:
        heartbeats = heartbeats[["start_sample", "end_sample", "r_peak_sample"]]
        heartbeats -= heartbeats["start_sample"].iloc[0]
    return heartbeats


def _get_heartbeat_borders(data: pd.DataFrame, heartbeats: pd.DataFrame) -> pd.DataFrame:
    start_samples = data.index[heartbeats["start_sample"]]
    end_sample = data.index[heartbeats["end_sample"] - 1][-1]
    # combine both into one array
    heartbeat_borders = start_samples.append(pd.Index([end_sample]))
    return heartbeat_borders



"""Example dataset for testing and demonstration purposes."""

__all__ = ["get_example_dataset"]

from pathlib import Path

from pepbench import __version__
from pepbench.datasets._example_dataset import ExampleDataset

LOCAL_EXAMPLE_PATH = Path(__file__).parent.parent.parent.joinpath("example_data")

PEPPI = None

if not LOCAL_EXAMPLE_PATH.exists():
    # if not (LOCAL_EXAMPLE_PATH / "README.md").is_file():
    import pooch

    GITHUB_FOLDER_PATH = "https://raw.githubusercontent.com/empkins/pepbench/{version}/example_data/"

    PEPPI = pooch.create(
        # Use the default cache folder for the operating system
        path=pooch.os_cache("pepbench"),
        # The remote data is on GitHub
        base_url=GITHUB_FOLDER_PATH,
        version=f"v{__version__}",
        version_dev="main",
        registry=None,
        # The name of an environment variable that *can* overwrite the path
        env="PEPBENCH_DATA_DIR",
    )

    # Get registry file from package_data
    # The registry file can be recreated by running the task `poe update_example_data`
    registry_file = LOCAL_EXAMPLE_PATH.joinpath("_example_data_registry.txt")
    # Load this registry file
    PEPPI.load_registry(registry_file)


def _pooch_get_folder(folder_path: Path) -> Path:
    """Get the path to the example data folder.

    If the data is not available locally, it will be downloaded from the remote repository.
    For this we use pooch to download all files that start with the folder name.
    """
    if PEPPI is None:
        return folder_path

    rel_folder_path = folder_path.relative_to(LOCAL_EXAMPLE_PATH)

    matching_files = []
    for f in PEPPI.registry:
        try:
            Path(f).relative_to(rel_folder_path)
        except ValueError:
            continue
        matching_files.append(Path(PEPPI.fetch(f, progressbar=True)))

    return PEPPI.abspath / rel_folder_path


def _pooch_get_file(file_path: Path) -> Path:
    """Get the path to the example data file.

    If the data is not available locally, it will be downloaded from the remote repository.
    For this we use pooch to download all files that start with the folder name.

    """
    if PEPPI is None:
        return file_path

    rel_folder_path = file_path.relative_to(LOCAL_EXAMPLE_PATH)

    return Path(PEPPI.fetch(str(rel_folder_path), progressbar=True))


def get_example_dataset(return_clean: bool = True) -> ExampleDataset:
    """Get an example dataset.

    Parameters
    ----------
    return_clean : bool, optional
        Whether to return cleaned/preprocessed signals when accessing the dataset or not. Default: True
        See the documentation of :class:`~pepbench.datasets.ExampleDataset` for more information.

    Returns
    -------
    :class:`~pepbench.datasets.ExampleDataset`
        An example dataset for testing and demonstration purposes.

    """
    fname = _pooch_get_file(LOCAL_EXAMPLE_PATH.joinpath("example_dataset.zip"))
    return ExampleDataset(example_file_path=fname, return_clean=return_clean)



"""Top-level module for the PEPBench package."""

from importlib import metadata

from pepbench import algorithms, datasets, evaluation, pipelines

__version__ = metadata.version(__name__)

__all__ = ["algorithms", "datasets", "evaluation", "pipelines"]



"""Evaluation challenges for systematic evaluation of PEP extraction pipelines.

The systematic evaluation of different PEP extraction pipelines requires a standardized evaluation procedure.
This can be achieved by defining a challenge that evaluates a pipeline on a given dataset and returns the results.
For that, `tpcp` provides the necessary functionality to define such challenges and compare the results of different
pipelines, implemented in :class:`pepbench.evaluation.PepEvaluationChallenge`.

The evaluation challenge takes a dataset
and a scoring function and evaluates the performance of a PEP extraction pipeline on the given dataset. While the
scoring function can be customized, the default scoring function is provided in
:func:`pepbench.evaluation.score_pep_evaluation`.

"""

from pepbench.evaluation._evaluation import ChallengeResults, PepEvaluationChallenge
from pepbench.evaluation._scoring import score_pep_evaluation

__all__ = ["ChallengeResults", "PepEvaluationChallenge", "score_pep_evaluation"]



"""Utility functions for data handling."""

import pandas as pd

__all__ = ["reindex_empkins", "reindex_guardian", "rename_empkins", "rename_guardian"]

condition_mapping_empkins = {"tsst": "TSST", "ftsst": "f-TSST"}
phase_mapping_empkins = {
    "Prep": "Preparation",
    "Pause_1": "Pause 1",
    "Talk": "Talk",
    "Math": "Math",
    "Pause_5": "Pause 5",
}
phase_mapping_guardian = {
    "Pause": "Pause",
    "Valsalva": "Valsalva",
    "HoldingBreath": "Apnea",
    "TiltUp": "Tilt-Up",
    "TiltDown": "Tilt-Down",
}


def reindex_empkins(data: pd.DataFrame, after_rename: bool = False) -> pd.DataFrame:
    """Reindex data from the *EmpkinSDataset*.

    The reindexing is performed according to the condition (tsst, ftsst) and phase (Prep, Pause_1, Talk, Math, Pause_5)
    mappings. The reindexing can be performed *before* or *after* the data has been renamed
    (using :func:`rename_empkins`).

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Data from the *EmpkinSDataset*.
    after_rename : bool, optional
        ``True`` if the data has already been renamed using :func:`rename_empkins`, ``False`` otherwise.
        Default: ``False``

    Returns
    -------
    :class:`pandas.DataFrame`
        Reindexed data from the *EmpkinSDataset*.

    """
    if after_rename:
        return data.reindex(condition_mapping_empkins.values(), level="condition").reindex(
            phase_mapping_empkins.values(), level="phase"
        )
    return data.reindex(condition_mapping_empkins.keys(), level="condition").reindex(
        phase_mapping_empkins.keys(), level="phase"
    )


def rename_empkins(data: pd.DataFrame) -> pd.DataFrame:
    """Rename the data from the *EmpkinSDataset*.

    The renaming is performed according to the condition (tsst -> TSST, ftsst -> f-TSST) and phase
    (Prep -> Preparation, Pause_1 -> Pause 1, Talk -> Talk, Math -> Math, Pause_5 -> Pause 5) mappings.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Data from the *EmpkinSDataset*.

    Returns
    -------
    :class:`pandas.DataFrame`
        Renamed data from the *EmpkinSDataset*.

    """
    return data.rename(condition_mapping_empkins, level="condition").rename(phase_mapping_empkins, level="phase")


def reindex_guardian(data: pd.DataFrame, after_rename: bool | None = False) -> pd.DataFrame:
    """Reindex data from the *GuardianDataset*.

    The reindexing is performed according to the phase (Pause, Valsalva, HoldingBreath, TiltUp, TiltDown) mappings.
    The reindexing can be performed *before* or *after* the data has been renamed (using :func:`rename_guardian`).

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Data from the *GuardianDataset*.
    after_rename : bool, optional
        ``True`` if the data has already been renamed using :func:`rename_guardian`, ``False`` otherwise.
        Default: ``False``

    Returns
    -------
    :class:`pandas.DataFrame`
        Reindexed data from the *GuardianDataset*.

    """
    if after_rename:
        return data.reindex(phase_mapping_guardian.values(), level="phase")
    return data.reindex(phase_mapping_guardian.keys(), level="phase")


def rename_guardian(data: pd.DataFrame) -> pd.DataFrame:
    """Rename the data from the *GuardianDataset*.

    The renaming is performed according to the phase (Pause -> Pause, Valsalva -> Valsalva, HoldingBreath -> Apnea,
    TiltUp -> Tilt-Up, TiltDown -> Tilt-Down) mappings.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Data from the *GuardianDataset*.

    Returns
    -------
    :class:`pandas.DataFrame`
        Renamed data from the *GuardianDataset*.

    """
    return data.rename(phase_mapping_guardian, level="phase")



"""Module for ICG outlier correction algorithms."""

from biopsykit.signals.icg.outlier_correction import (
    OutlierCorrectionDummy,
    OutlierCorrectionForouzanfar2018,
    OutlierCorrectionLinearInterpolation,
)

__all__ = ["OutlierCorrectionDummy", "OutlierCorrectionForouzanfar2018", "OutlierCorrectionLinearInterpolation"]
